package state51::Logger;
use MooseX::Singleton;
use Class::MOP ();
use Class::Load;
use MooseX::Types::Moose qw/ Int Str Bool ArrayRef HashRef /;
use Carp qw/croak cluck/;
use Sys::Hostname ();
use namespace::autoclean;

our $VERSION = '2022052300';

our @BACKENDS;
BEGIN {
    $ENV{PLATFORM_LOG_BACKENDS} ||= 'Console';
    @BACKENDS = map { 'state51::Logger::' . $_ } split(/,/, $ENV{PLATFORM_LOG_BACKENDS});
    foreach my $class (@BACKENDS) {
        Class::Load::load_class($class);
    }
}
=head1 NAME

state51::Logger

=head1 DESCRIPTION

    A simple class for keeping logs.

    my $log = state51::Logger->new(trace   => 1);

    $log->message(['ordinary message']);


    Messages are always arrays.
    There are four levels of message, "warn", "fatal", "info" and "trace";

    All messages go to the log file, timestamped and with PID.

    "info" will also be printed on STDOUT, "warn" and "fatal" will go to STDERR.

    "trace" will go to STDERR, but only if the log object has been initialised with a true value for trace.

    "fatal" message will cause a fatal croak (after the message + croak message have been logged)

=cut

our %log_levels = ( 'fatal' => 1,
                    'error' => 1,
                    'warn'  => 1,
                    'info'  => 1,
                    'trace' => 1,
                    'debug' => 1,
                    'billable' => 1,
                );

# this shouldn't ever be called by as_string, but it does
sub FREEZE {}

# args
# trace   => "send trace messages to STDERR"
has pid => (
    isa => Int,
    is => 'ro',
    init_arg => undef,
    default => $$,
);

has hostname => (
    isa => Str,
    is => 'ro',
    init_arg => undef,
    default => sub { (split('\.',Sys::Hostname::hostname()))[0] },
);

has cmd => (
    isa => Str,
    is => 'ro',
    init_arg => undef,
    default => sub {
        my $cmd = (split('/',$0))[-1];
        substr $cmd, 0, 39; # Needs to fit into 'website' in scribe, which is
                            # a varchar(40)
    },
);

has trace => (
    isa => Bool,
    is => 'rw',
    default => sub {
        return $ENV{LOGGER_TRACE};
    },
);

has backends => (
    isa => ArrayRef,
    is => 'ro',
    lazy => 1,
    default => sub {
        my $self = shift;
        return [
            map { $_->new(trace => $self->trace) } @BACKENDS
        ];
    },
);

has backends_by_name => (
    isa => HashRef,
    is => 'ro',
    lazy => 1,
    default => sub {
        return { map { ref($_) => $_ } @{ shift->backends } };
    },
);

sub BUILD { shift->backends }
__PACKAGE__->initialize; # Cause singleton instance to build up front..
my $instance = __PACKAGE__->instance;
override new => sub { $instance };

sub trace_on { shift->trace(1) }

sub _message_args {
    my ($self, @args) = @_;

    my %args;
    if ( (scalar(@args) == 1 || scalar(@args) == 2) && ! ref($args[0])) {
        $args[0] = [ $args[0] ];
    }
    if (ref($args[0]) eq 'ARRAY') {
        $args{level} = $args[1] || 'info'; # default level
        $args{message} = $args[0];
    }
    else {
        %args = @args;

        if ($args{level}) {
            croak "unknown error level <<".$args{level}.">>" unless $log_levels{$args{level}};
        }
        else {
            $args{level} = 'info';
        }

        unless ($args{message}) {
            $args{level} = 'warn';
            $args{message} = ['missing message parameter'];
            Carp::cluck('missing message parameter');
        }
    }
    return %args;
}

sub message {
    my $self = shift;
    my %args = $self->_message_args(@_);
    my @receipts;

    foreach my $b (@{ $self->backends }) {
        my $receipt = $b->message($self, \%args);
        push @receipts, @{$receipt} if ref $receipt eq 'ARRAY';
    }
    if ($args{level} eq 'fatal') {
        croak("Fatal: " . join(", ", @{$args{message}}));
    }

    return \@receipts;
}

# Note that we may be idle for a while, this allows any backends which have persistent connections
# or whatever to close them..
sub idle {
    my $self = shift;
    foreach my $b (@{ $self->backends }) {
        my $code = $b->can('idle');
        $b->$code($self) if $code;
    }
}

sub block_till_all_messages_sent {
    my $self = shift;
    foreach my $b (@{ $self->backends }) {
        my $code = $b->can('block_till_all_messages_sent');
        $b->$code($self) if $code;
    }
}

# FIXME - Remove this, done automatically now!
sub pre_fork {}

1;

